/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 1999-2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */


#include <linux/config.h>

#if defined(CONFIG_SMP) && !defined(__SMP__)
#define __SMP__
#endif


// pull in the versioned kernel api defines if needed
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>

#include "nv-linux.h"			        // needs to precede other headers (SMP)
#define RM_HEADERS_MINIMAL      // we just need top half of nvrm.h, not the entire
#include <nvrm.h>               // tree of include files
#include <nv_compiler.h>         // get the compiler version used

#include <linux/stddef.h>
#include <linux/mm.h>
#include <linux/kernel.h> 		// printk()
#include <linux/fs.h>     		// everything...
#include <linux/errno.h>  		// error codes
#include <linux/types.h>  		// size_t
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        	// O_ACCMODE
#include <linux/pci.h>    		// PCI stuff
#include <linux/delay.h>  		// udelay()
#include <linux/interrupt.h>
#include <linux/tqueue.h>               // struct tq_struct
#include <linux/poll.h>
#ifdef CONFIG_PM
#include <linux/pm.h>                   // power management
#endif

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#endif

#include <asm/page.h>
#include <asm/pgtable.h>  		// pte bit definitions
#include <asm/system.h>                 // cli(), *_flags
#include <asm/segment.h>                // memcpy and verify_write
#include <asm/uaccess.h>		// access_ok
#include <asm/io.h>
#include <asm/softirq.h>                // mark_bh
#if !defined (KERNEL_2_2)
#include <linux/spinlock.h>
#else
#include <asm/spinlock.h>
#endif
#include <asm/smp.h>

#include <linux/wrapper.h>

#include "nv_ref.h"

#include "os-interface.h"              // KernInitAgp

#include "nv-ids.h"

#define LINUX_VMA_DEV(vma)  ((vma)->vm_file->f_dentry->d_inode->i_rdev)

#if !defined (KERNEL_2_2)
#define LINUX_VMA_OFFS(vma) (((vma)->vm_pgoff) << PAGE_SHIFT)
#else
#define LINUX_VMA_OFFS(vma) ((vma)->vm_offset)
#endif

#define NV_PAGE_ALIGN(addr)    ( ((addr) + PAGE_SIZE - 1) / PAGE_SIZE)
#define NV_MASK_OFFSET(addr)  ( (addr) & (PAGE_SIZE - 1) )

#ifndef NV01_ROOT
#define NV01_ROOT 0x00000000
#endif

/*
 * our global state; one per device
 */

nv_linux_state_t nv_linux_devices[NV_MAX_DEVICES] = { { { 0 } } };

#ifdef CONFIG_PM
/* XXX PM do we only need one, or one for each device? */
static struct pm_dev *pm_nv_dev;
#endif

/*
 * And one for the control device
 */

nv_linux_state_t nv_ctl_device = { { 0 } };

// keep track of opened clients and their process id so they
//   can be free'd up on abnormal close
nv_client_t       nv_clients[NV_MAX_CLIENTS];
struct tq_struct  nv_bottom_halves[NV_MAX_CLIENTS];


#ifdef CONFIG_DEVFS_FS
devfs_handle_t  nv_dev_handle[NV_MAX_DEVICES];
devfs_handle_t  nv_ctl_handle;
#endif

/*
 * pick apart our minor device number
 * low 3 bits is NV device
 * if 255, then its the control device
 */

#define NV_DEVICE_NUMBER(_minor) ((_minor) & 0x0f)
#define NV_DEVICE_IS_CONTROL_DEVICE(_minor) \
             (((_minor) & 0xFF) == 0xFF)

/*
 * Hardware access macros for the kernel driver only
 * The resource manager part of the driver uses different values here
 */

// HW register access macros
#define REG_WR08(a,d)   nv->regs.map_u->Reg008[(a)]   = (U008) (d)
#define REG_WR16(a,d)   nv->regs.map_u->Reg016[(a)/2] = (U016) (d)
#define REG_WR32(a,d)   nv->regs.map_u->Reg032[(a)/4] = (U032) (d)

#define REG_RD08(a)     nv->regs.map_u->Reg008[(a)]
#define REG_RD16(a)     nv->regs.map_u->Reg016[(a)/2]
#define REG_RD32(a)     nv->regs.map_u->Reg032[(a)/4]



#ifdef DEBUG
#define NV_DMSG(nv, fmt, args...)  NV_iDMSG(nv, fmt, ## args)
#else
#define NV_DMSG(nv, fmt, args...)
#endif


/* functions in core resman that we need to call */
void NvUpdateAGPConfig(PHWINFO);
BOOL RmInitAGP(PHWINFO);
BOOL RmTeardownAGP(PHWINFO);
RM_STATUS NvAllocAGPPages ( PHWINFO, VOID **, U032, U032, VOID **);
RM_STATUS NvMapAGPPages ( PHWINFO, VOID **, U032, VOID *);
RM_STATUS NvFreeAGPPages ( PHWINFO, VOID **, VOID * );

/***
 *** STATIC functions, only in this file
 ***/

/* nvos_ functions.. do not take a state device parameter  */
static int      nvos_is_nv_device(struct pci_dev *dev);
static int      nvos_set_primary_card(nv_ioctl_primary_card_t *info);
static int      nvos_probe_devices(void);
static void *   nvos_malloc(unsigned long);
static void     nvos_free(void **);

static void     nvos_proc_create(void);
static void     nvos_proc_remove(void);

/* nvl_ functions.. take a linux state device pointer */
static nv_alloc_t   *nvl_find_alloc(nv_linux_state_t *, unsigned long, nv_alloc_t **);
static nv_alloc_t   *nvl_find_agp_alloc(nv_linux_state_t *, unsigned long, nv_alloc_t **);
static nv_alloc_t   *nvl_find_alloc_by_num_pages(nv_linux_state_t *, unsigned long, nv_alloc_t **);

#if defined(IA64) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 9))
static nv_alloc_t   *nvl_find_agp_alloc_by_vma(nv_linux_state_t *nvl, unsigned long vaddr);
#endif

/* nv_ functions. not used in main resman, but take general state device */
static void          nv_interrupts_disable(nv_state_t *);
static void          nv_interrupts_enable(nv_state_t *);
static int           nv_map_device(nv_state_t *);
static void          nv_unmap_device(nv_state_t *);

/* lock-related functions that should only be called from this file */
static void nv_lock_init_locks(nv_state_t *nv);
static void nv_lock_ldata(nv_state_t *nv);
static void nv_unlock_ldata(nv_state_t *nv);
static void nv_lock_at(nv_state_t *nv);
static void nv_unlock_at(nv_state_t *nv);

/***
 *** EXPORTS to Linux Kernel
 ***/

/* linux module interface functions (called by linux kernel) */
int           init_module(void);
void          cleanup_module(void);

/* nv_kern_ functions, interfaces used by linux kernel */
void          nv_kern_vma_open(struct vm_area_struct *vma);
void          nv_kern_vma_release(struct vm_area_struct *vma);

int           nv_kern_open(struct inode *, struct file *);
int           nv_kern_close(struct inode *, struct file *);
int           nv_kern_mmap(struct file *, struct vm_area_struct *);
unsigned int  nv_kern_poll(struct file *, poll_table *);
int           nv_kern_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
void          nv_kern_isr(int, void *, struct pt_regs *);
#ifdef CONFIG_PM
int           nv_kern_pm(struct pm_dev *dev, pm_request_t rqst, void *data);
#endif

int           nv_kern_read_cardinfo(char *, char **, off_t off, int, int *, void *);
int           nv_kern_read_status(char *, char **, off_t off, int, int *, void *);
int           nv_kern_read_agpinfo(char *, char **, off_t off, int, int *, void *);
int           nv_kern_read_version(char *, char **, off_t off, int, int *, void *);
int           nv_kern_read_legacy(char *, char **, off_t off, int, int *, void *);

int           nv_kern_ctl_open(struct inode *, struct file *);
int           nv_kern_ctl_close(struct inode *, struct file *);
unsigned int  nv_kern_ctl_poll(struct file *, poll_table *);

/***
 *** see nv.h for functions exported to other parts of resman
 ***/


/* character driver entry points */

static struct file_operations nv_fops = {
    poll:           nv_kern_poll,
    ioctl:          nv_kern_ioctl,
    mmap:           nv_kern_mmap,
    open:           nv_kern_open,
    release:        nv_kern_close,
};

// Our reserved major device number.
int nv_major = NV_MAJOR_DEVICE_NUMBER;

// pull in the pointer to the NVID stamp from the binary module
extern const char *pNVRM_ID;

// do not call this directly
void NV_iMSG(
    nv_state_t *nv,
    const char *kernel_message_level,
    const char *printf_format,
    va_list arglist
  )
{
    char nv_error_string[256];
    char *p;

    p = nv_error_string;

    if (kernel_message_level)
        p += sprintf(p, "%s", kernel_message_level);

    if (nv)
        if (nv->device_number == NV_CONTROL_DEVICE_NUMBER)
            p += sprintf(p, "NVctl: ");
        else
            p += sprintf(p, "NV%d: ", nv->device_number);
    else
        p += sprintf(p, "NV: ");

    p += vsprintf(p, printf_format, arglist);
    printk("%s\n", nv_error_string);
}

// should only be called using NV_DMSG macro
void NV_iDMSG(
    nv_state_t *nv,
    const char *printf_format,
    ...
)
{
    va_list arglist;

    va_start(arglist, printf_format);
    NV_iMSG(nv, KERN_INFO, printf_format, arglist);
    va_end(arglist);
}

void NV_EMSG(
    nv_state_t *nv,
    const char *printf_format,
    ...
  )
{
    va_list arglist;

    va_start(arglist, printf_format);
    NV_iMSG(nv, KERN_ERR, printf_format, arglist);
    va_end(arglist);
}



/***
 *** STATIC functions
 ***/

/* set which card is primary */
static int nvos_set_primary_card(nv_ioctl_primary_card_t *info)
{
    nv_linux_devices[info->index].nv_state.flags |= NV_FLAG_POSTED;

    return 0;
}

/* return 1 if device is ours, 0 otherwise */
static int nvos_is_nv_device(struct pci_dev *dev)
{
    int i;

    /* check for our vendor id */
    if (dev->vendor != 0x10de)
        return 0;

    /* sChipInfoChar and CHIP_INFO_CHAR is defined by nv-ids.h */
    for (i = 0; i < sizeof(sChipInfoChar) / sizeof(CHIP_INFO_CHAR); i++) {
        if (sChipInfoChar[i].ulDevID == dev->device) {
            return 1;
        }
    }

    printk ("nvidia: %04x:%04x - unknown PCI id\n",
            dev->vendor, dev->device);

    return 0;
}

/* get pci aperture information */
static void
nvos_get_pci_size(struct pci_dev *dev, nv_aperture_t *ap, u32 base)
{
    u32 base_reg, size;

    pci_read_config_dword(dev, base, &base_reg);
    pci_write_config_dword(dev, base, ~0);
    pci_read_config_dword(dev, base, &size);
    pci_write_config_dword(dev, base, base_reg);

    if (base == PCI_BASE_ADDRESS_0) {
        /* registers */
        ap->address = base_reg & PCI_BASE_ADDRESS_IO_MASK;
        size &= PCI_BASE_ADDRESS_IO_MASK;
    } else {
        /* framebuffer */
        ap->address = base_reg & PCI_BASE_ADDRESS_MEM_MASK;
        size &= PCI_BASE_ADDRESS_MEM_MASK;
    }

    /* translate the size */
    ap->size = ~(size - 1) & 0xffffffff;
}

/* find nvidia devices and set initial state */
static int
nvos_probe_devices(void)
{
    unsigned short count;
    unsigned short cmd;
    struct pci_dev *dev;

    /* for state tracking */
    nv_state_t *nv;
    nv_linux_state_t *nvl;

    count = 0;
    dev = (struct pci_dev *) 0;

    dev = pci_find_class(PCI_CLASS_DISPLAY_VGA << 8, dev);
    do {
        if (!nvos_is_nv_device(dev))
            goto next;

        /* initialize bus-dependent config state */
        nvl = &nv_linux_devices[count];
        nv  = NV_STATE_PTR(nvl);

        nv->vendor_id     = dev->vendor;
        nv->device_id     = dev->device;
        nv->os_state      = (void *) nvl;
        nv->bus           = dev->bus->number;
        nv->slot          = PCI_SLOT(dev->devfn);
        nv->device_number = count;

        pci_read_config_word(dev, PCI_COMMAND, &cmd);
        cmd &= PCI_COMMAND_MEMORY | (cmd & PCI_COMMAND_MASTER);

        nv_lock_init_locks(nv);

        /* is the device currently enabled? */
        if (cmd == 0) {
            NV_DMSG(nv, "device not enabled; enabling");
            cmd |= PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER;
            pci_write_config_word(dev, PCI_COMMAND, cmd);
        }
    
        nvos_get_pci_size(dev, &nv->regs, PCI_BASE_ADDRESS_0);
        nvos_get_pci_size(dev, &nv->fb,   PCI_BASE_ADDRESS_1);
    
        nv->interrupt_line = dev->irq;

        /* check common error condition */
        if (nv->interrupt_line == 0) {
            printk("nvidia: Can't find an IRQ for your NVIDIA card!  \n");
            printk("nvidia: Please check your BIOS settings.         \n");
            printk("nvidia: [Plug & Play OS   ] should be set to NO  \n");
            printk("nvidia: [Assign IRQ to VGA] should be set to YES \n");
        }

        NV_DMSG(nv, "%02x:%02x.%x %04x:%04x - 0x%08x [size=%dM]",
                nv->bus, nv->slot, PCI_FUNC(dev->devfn),
                nv->vendor_id, nv->device_id, nv->regs.address,
                nv->regs.size / (1024 * 1024));
        NV_DMSG(nv, "%02x:%02x.%x %04x:%04x - 0x%08x [size=%dM]",
                nv->bus, nv->slot, PCI_FUNC(dev->devfn),
                nv->vendor_id, nv->device_id, nv->fb.address,
                nv->fb.size / (1024 * 1024));

        if (count++ == NV_MAX_DEVICES - 1) {
            NV_DMSG(nv, "maximum device number (%d) reached!", count);
            break;
        }

    next:
        dev = pci_find_class(PCI_CLASS_DISPLAY_VGA << 8, dev);
    } while (dev);

    NV_DMSG(NULL, "found %d device%s", count, count ? "" : "s");

    return count;
}

/*
 * The idea behind nvos_malloc is to manage physical memory for DMA buffers
 * directly (as opposed to using "vmalloc") to avoid exhausting the kernel's
 * virtual address space which can shrink down to 128MB on i386 systems with
 * large amounts of physical memory. This also minimizes fragmentation of
 * virtual memory.
 * While this isn't really necessary on ia64, it can't hurt either.
 */

/*
 * When called with __GFP_DMA, __get_free_pages should return physical pages
 * addressable by 32bit PCI hardware on the ia64. On ia32, this will always
 * be the case unless we explicitly tell __get_free_pages to give us "high"
 * memory on systems with high memory support enabled.
 */

#if defined(IA64)
#define NV_GFP_HW (GFP_KERNEL | __GFP_DMA)
#else
#define NV_GFP_HW (GFP_KERNEL)
#endif
 
static void *nvos_malloc(unsigned long size)
{
    unsigned long *page_list = NULL;
    unsigned long *page_ptr  = NULL;
    unsigned int   pages_needed;
    unsigned int   page_list_size;

    /*
     * allocate a pointer for each physical page and an
     * integer to hold the number of pages allocated
     */
    pages_needed = (size >> PAGE_SHIFT);
    page_list_size = (pages_needed + 1) * sizeof(unsigned long *);

    page_list = vmalloc(page_list_size);
    if (page_list == NULL)
        return NULL;

    /* store the number of physical pages allocated */
    *page_list++ = (unsigned long) pages_needed;

    /* point page_ptr at the start of the actual page list */
    page_ptr = page_list;

    while (pages_needed) {
        unsigned long virt_addr;
        unsigned long phys_addr;
        
        virt_addr = __get_free_pages(NV_GFP_HW, 0);
        if (virt_addr == 0) {
            goto failed;
        }
        phys_addr = virt_to_phys((void *) virt_addr);

        /* lock the page for dma purposes */
        mem_map_reserve(GET_MAP_NR(phys_addr));

        *page_ptr++ = phys_addr;
        pages_needed--;
    }

    return page_list;

failed:
    /* back up to last valid page */
    page_ptr -= 1;

    while (page_ptr != page_list) {
        mem_map_unreserve(GET_MAP_NR(*page_ptr));
        free_page((unsigned long) phys_to_virt(*page_ptr));
        page_ptr--;
    }

    /* back up over number of pages allocated */
    page_list -= 1;
    vfree(page_list);

    return NULL;
}

static void nvos_free(void **page_list)
{
    unsigned long *page_ptr;
    unsigned int   pages_left;

    page_ptr = (unsigned long *) page_list;

    /* retrieve the number of pages allocated */
    pages_left = *(unsigned int *) --page_list;

    while (pages_left) {
        mem_map_unreserve(GET_MAP_NR(*page_ptr));
        free_page((unsigned long) phys_to_virt(*page_ptr++));
        pages_left--;
    }

    vfree(page_list);
}

static u8 nvos_find_agp_capability(struct pci_dev *dev)
{
    u16 status;
    u8  cap_ptr, cap_id;

    pci_read_config_word(dev, PCI_STATUS, &status);
    status &= PCI_STATUS_CAP_LIST;
    if (!status)
        return 0;

    switch (dev->hdr_type) {
        case PCI_HEADER_TYPE_NORMAL:
        case PCI_HEADER_TYPE_BRIDGE:
            pci_read_config_byte(dev, PCI_CAPABILITY_LIST, &cap_ptr);
            break;
        default:
            return 0;
    }

    do {
        cap_ptr &= 0xfc;
        pci_read_config_byte(dev, cap_ptr + PCI_CAP_LIST_ID, &cap_id);
        if (cap_id == PCI_CAP_ID_AGP)
            return cap_ptr;
        pci_read_config_byte(dev, cap_ptr + PCI_CAP_LIST_NEXT, &cap_ptr);
    } while (cap_ptr && cap_id != 0xff);

    return 0;
}

static struct pci_dev* nvos_find_agp_by_class(unsigned int class)
{
    struct pci_dev *dev;
    u8     cap_ptr;

    dev = pci_find_class(class << 8, NULL);
    do {
        cap_ptr = nvos_find_agp_capability(dev);
        if (cap_ptr)
            return dev;
        dev = pci_find_class(class << 8, dev);
    } while (dev);

    return NULL;
}

static struct pci_dev* nv_find_pci_dev(nv_state_t *nv)
{
    struct pci_dev *dev;
    int    bus  = 0;
    int    slot = 0;
    
    dev = NULL;
    do {
        dev = pci_find_device(nv->vendor_id, nv->device_id, dev);
        if (dev) {
            bus = dev->bus->number;
            slot = PCI_SLOT(dev->devfn);
        }
    } while (dev && (bus != nv->bus || slot != nv->slot));

    return dev;
}

static void nvos_proc_create(void)
{
#ifdef CONFIG_PROC_FS
    struct pci_dev *dev;
    int i = 0;
    char name[6];

    struct proc_dir_entry *entry;
    struct proc_dir_entry *proc[5];

    /* world readable directory */
    int flags = S_IFDIR | S_IRUGO | S_IXUGO;

    enum { NVIDIA, DRIVER, CARDS, AGP, LEGACY };

    nv_state_t *nv;
    nv_linux_state_t *nvl;
    nv_linux_state_t *nv_max_devices;

#if defined (KERNEL_2_2)
    proc[DRIVER] = create_proc_entry("driver", flags, &proc_root);
#else
    proc[DRIVER] = proc_root_driver;
#endif

    proc[NVIDIA] = create_proc_entry("nvidia", flags, proc[DRIVER]);
    proc[AGP]    = create_proc_entry("agp",    flags, proc[NVIDIA]);
    proc[CARDS]  = create_proc_entry("cards",  flags, proc[NVIDIA]);
    proc[LEGACY] = create_proc_entry("nv",     flags, &proc_root);

    nv_max_devices = nv_linux_devices + NV_MAX_DEVICES;
    for (nvl = nv_linux_devices; nvl < nv_max_devices; nvl++) 
    {
        nv = NV_STATE_PTR(nvl);

        if (nv->device_id == 0)
            break;

        /* world readable file */
        flags = S_IFREG | S_IRUGO;

        dev = nv_find_pci_dev(nv);
        if (!dev)
            break;

        sprintf(name, "card%d", i);
        entry = create_proc_entry(name, flags, proc[LEGACY]);
        entry->read_proc = nv_kern_read_legacy;
        
        sprintf(name, "%d", i++);
        entry = create_proc_entry(name, flags, proc[CARDS]);
        entry->data = nv;
        entry->read_proc = nv_kern_read_cardinfo;

        if (nvos_find_agp_capability(dev)) {
            entry = create_proc_entry("status", flags, proc[AGP]);
            entry->data = nv;
            entry->read_proc = nv_kern_read_status;

            entry = create_proc_entry("card", flags, proc[AGP]);
            entry->data = nv;
            entry->read_proc = nv_kern_read_agpinfo;
        }
    }

    entry = create_proc_entry("version", flags, proc[NVIDIA]);
    entry->read_proc = nv_kern_read_version;

    entry = create_proc_entry("host-bridge", flags, proc[AGP]);
    entry->data = NULL;
    entry->read_proc = nv_kern_read_agpinfo;
#endif
}

static void nvos_proc_remove(void)
{
#ifdef CONFIG_PROC_FS
#if defined (KERNEL_2_2)
    remove_proc_entry("driver", &proc_root);
    remove_proc_entry("nv", &proc_root);
#else
    remove_proc_entry("nvidia", proc_root_driver);
    remove_proc_entry("nv", &proc_root);
#endif
#endif
}

/*
 * Given a virtual address, fid the 'at' that owns it
 * Uses the physical address as the key.
 */
static nv_alloc_t *nvl_find_alloc(
    nv_linux_state_t    *nvl,
    unsigned long  address,
    nv_alloc_t   **prev
)
{
    nv_alloc_t *dummy_prev;
    nv_alloc_t *at;
    void *physaddr, *offset;
    unsigned int i;

    address = address & PAGE_MASK;
    physaddr = (void *) nv_get_phys_address(address);

    if (physaddr == NULL)
        goto done;

    if (prev == (nv_alloc_t **) 0)
        prev = &dummy_prev;

    *prev = (nv_alloc_t *) 0;
    for (at = nvl->alloc_queue; at; *prev = at, at = at->next)
    {
        if (at->page_table)
        {
            /* alloc/free pages will be sending the page_table address */
            if (address == ((unsigned long) at->page_table & PAGE_MASK))
                return at;

            for (i = 0; i < at->num_pages; i++)
            {
                offset = at->page_table[i];
                if ( (physaddr >= offset) &&
                     ((unsigned long)physaddr < ((unsigned long)offset + PAGE_SIZE - 1)) )
                    return at;
            }
        }
        else if (at->class == NV01_ROOT)
        {
            if (address == (unsigned long) at->direct_kmapping)
                return at;
        }
    }

done:
    /* failure is not necessarily an error if the caller
       was just probing an address */
    NV_DMSG(NV_STATE_PTR(nvl), "could not find map for vm 0x%lx", address);
    return (nv_alloc_t *) 0;
}

/* Given a kernel virtual address, find the 'at' that owns it */
static nv_alloc_t *nvl_find_agp_alloc(
    nv_linux_state_t    *nvl,
    unsigned long  address,
    nv_alloc_t   **prev
)
{
    nv_alloc_t *dummy_prev;
    nv_alloc_t *at;

    if (prev == (nv_alloc_t **) 0)
        prev = &dummy_prev;

    *prev = (nv_alloc_t *) 0;
    for (at = nvl->alloc_queue; at; *prev = at, at = at->next)
    {
        if (NV_OSAGP_ENABLED(NV_STATE_PTR(nvl)))
        {
            if (at->agp_mapping == NULL)
                continue;
            if (address == (unsigned long) at->agp_mapping)
                return at;
        } else {
            if ((void *) address == (void *) at->page_table)
                return at;
        }
    }

    /* failure is not necessarily an error if the caller
       was just probing an address */
    NV_DMSG(NV_STATE_PTR(nvl), "could not find map for vm 0x%lx", address);
    return (nv_alloc_t *) 0;
}

#if defined(IA64) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 9))
/*
 * Given a virtual address, find the 'at' that owns it. Make sure we
 * are only looking at the current thread group. (nv_kern_vma_nopage)
 */
static nv_alloc_t *
nvl_find_agp_alloc_by_vma(nv_linux_state_t *nvl, unsigned long vaddr)
{
    nv_alloc_t *at;

    for (at = nvl->alloc_queue; at; at = at->next)
    {
        if (at->agp_mapping == NULL)
            continue;
        if (at->thread_gid != current->tgid)
            continue;
        if (at->vma->vm_start == vaddr)
            return at;
    }
    /* failure is an error here */
    return (nv_alloc_t *) 0;
}
#endif /* defined (IA64) */

/*
 * We allocated system/agp memory in an ioctl before a mmap, and now we need to
 * retreive it. Find the allocation based on number of pages and having no vma
 * associated with the address.
 * XXX getting 2 allocations mixed up at this point shouldn't be a problem,
 * since the memory isn't associated with a client/user yet.
 */
static nv_alloc_t *nvl_find_alloc_by_num_pages(
    nv_linux_state_t    *nvl,
    unsigned long  num_pages,
    nv_alloc_t   **prev
)
{
    nv_alloc_t *dummy_prev;
    nv_alloc_t *at;

    if (prev == (nv_alloc_t **) 0)
        prev = &dummy_prev;

    *prev = (nv_alloc_t *) 0;
    for (at = nvl->alloc_queue; at; *prev = at, at = at->next)
    {
        if ((at->num_pages == num_pages) &&
            (at->process_id == current->pid) &&
            (at->class != NV01_ROOT) &&
            (at->vma == NULL) )
            return at;
    }

    /* failure is not necessarily an error if the caller
       was just probing an address */
    NV_DMSG(NV_STATE_PTR(nvl), "could not find a mapping with %d pages",
        num_pages);

    return (nv_alloc_t *) 0;
}


static void nv_interrupts_disable(
    nv_state_t *nv
)
{
    nv->regs.map[NV_PMC_INTR_EN_0 / 4] = 0;
}

static void nv_interrupts_enable(
    nv_state_t *nv
)
{
    nv->regs.map[NV_PMC_INTR_EN_0 / 4] = 1;
}


/*
 * if the reported aperture is larger than the framebuffer, it will have
 * duplicate mappings of it (e.g. base + 0, base + fb size, ...). We can
 * use this to detect the real size of the framebuffer when PCI config
 * space misreports it. This allows us to minimize the amount of virtual
 * address space used.
 */

static int nv_map_device(nv_state_t *nv)
{
    unsigned long base;
    unsigned int  size;

    nv->fb.map   = NULL;
    nv->fb.map_u = NULL;

    nv->regs.map   = NULL;
    nv->regs.map_u = NULL;
    
    base = nv->regs.address;
    size = nv->regs.size;
    
    /* map the registers */
    nv->regs.map = ioremap_nocache(base, size);

    if (!nv->regs.map || nv->regs.map == (void *) base) {
        iounmap(nv->fb.map);
        nv->fb.map   = NULL;
        nv->fb.map_u = NULL;
        printk("nvidia: couldn't map registers!\n");
        return -1;
    }

    nv->regs.map_u = (nv_phwreg_t) nv->regs.map;
    nv->fb.map_u   = (nv_phwreg_t) nv->fb.map;

    NV_DMSG(nv, "%dM rg @ 0x%x mapped to 0x%p",
            size / (1024 * 1024),
            nv->regs.address, nv->regs.map);

    return 0;
}

static void nv_unmap_device(nv_state_t *nv)
{
    iounmap((void *) nv->regs.map);
    nv->regs.map   = NULL;
    nv->regs.map_u = NULL;

    if (nv->fb.map) {
        iounmap((void *) nv->fb.map);
        nv->fb.map   = NULL;
        nv->fb.map_u = NULL;
    }
}


/***
 *** EXPORTS to Linux Kernel
 ***/

int init_module(void)
{
    nv_linux_state_t *nvl;
    int rc;
    int num_devices;

    memset(nv_linux_devices, 0, sizeof(nv_linux_devices));
    num_devices = nvos_probe_devices();

    if (num_devices == 0) {
        printk("nvidia: no NVIDIA graphics adapter found\n");
        return -ENODEV;
    }

    printk("nvidia: loading %s\n", pNVRM_ID);

#ifdef CONFIG_DEVFS_FS
    rc = devfs_register_chrdev(nv_major, "nvidia", &nv_fops);
#else
    rc = register_chrdev(nv_major, "nvidia", &nv_fops);
#endif

    if (rc < 0) {
        NV_EMSG((nv_state_t *) 0, "init_module: register failed");
        return rc;
    }

#ifdef CONFIG_DEVFS_FS
    osMemSet(nv_dev_handle, 0, sizeof(devfs_handle_t) * NV_MAX_DEVICES);
    do {
        char name[10];
        int i;

        nv_ctl_handle = devfs_register(NULL, "nvidiactl",
                            DEVFS_FL_DEFAULT, nv_major, 255,
                            S_IFCHR | S_IRUGO | S_IWUGO,
                            &nv_fops, NULL);

        for (i = 0; i < num_devices; i++) {
            sprintf(name, "nvidia%d", i);
            nv_dev_handle[i] = devfs_register(NULL, name,
                                  DEVFS_FL_DEFAULT, nv_major, i,
                                  S_IFCHR | S_IRUGO | S_IWUGO,
                                  &nv_fops, NULL);
        }
    } while(0);
#endif

    NV_DMSG((nv_state_t *) 0, "init_module: major number %d", nv_major);

    // init all the bottom half structures
    for (nvl = nv_linux_devices; nvl < nv_linux_devices + NV_MAX_DEVICES; nvl++)
    {
        nvl->bh = &nv_bottom_halves[nvl - nv_linux_devices];
        nvl->bh->routine = RmIsrBottomHalf;
        nvl->bh->data = (void *) 0;
        nvl->bh->sync = 0;
    }

    // init the control device
    {
        nv_state_t *nv_ctl = NV_STATE_PTR(&nv_ctl_device);
        nv_ctl_device.event_queue = NULL;
        nv_ctl->os_state = &nv_ctl_device;
        nv_lock_init_locks(nv_ctl);
    }

#ifdef CONFIG_PM
    /* XXX PM egads, is this the right place to do this? */
    //printk("registering Power Management\n");
    pm_nv_dev = pm_register(PM_PCI_DEV, PM_SYS_VGA, nv_kern_pm);
#endif

    // Init the resource manager
    if ( ! RmInitRm())
    {
        NV_EMSG((nv_state_t *) 0, "RmInitRm() failed");
        rc = -EIO;
        goto failed;
    }

    nvos_proc_create();

    return 0;

 failed:
#ifdef CONFIG_DEVFS_FS
    devfs_unregister_chrdev(nv_major, "nvidia");
#else
    unregister_chrdev(nv_major, "nvidia");
#endif
    return rc;
}

void cleanup_module(void)
{
    int rc;
    nv_linux_state_t *nvl;
    nv_linux_state_t *max_devices;

    /* remove /proc/driver/nvidia */
    nvos_proc_remove();

    NV_DMSG((nv_state_t *) 0, "cleanup_module");

#ifdef CONFIG_PM
    /* XXX PM egads, is this the right place to do this? */
    //printk("unregistering Power Management\n");
    pm_unregister(pm_nv_dev);
#endif

    // Shutdown the resource manager
    RmShutdownRm();

    /*
     * Make sure we freed up all the mappings. The kernel should
     * do this automatically before calling close.
     */
    max_devices = nv_linux_devices + NV_MAX_DEVICES;
    for (nvl = nv_linux_devices; nvl < max_devices; nvl++)
    {
        if (nvl->alloc_queue)
        {
            if (nvl->alloc_queue->vma == NULL)
                continue;

            NV_EMSG(NV_STATE_PTR(nvl),
                    "still have vm que at cleanup_module(): 0x%x to 0x%x",
                    nvl->alloc_queue->vma->vm_start,
                    nvl->alloc_queue->vma->vm_end);
        }
    }

#ifdef CONFIG_DEVFS_FS
    rc = devfs_unregister_chrdev(nv_major, "nvidia");
#else
    rc = unregister_chrdev(nv_major, "nvidia");
#endif

    if (rc < 0) {
        NV_EMSG((nv_state_t *) 0, "cleanup_module: unregister nv failed");
    }

#ifdef CONFIG_DEVFS_FS
    do {
        int i;
        for (i = 0; nv_dev_handle[i] != 0; i++) {
            devfs_unregister(nv_dev_handle[i]);
        }
    } while(0);
    devfs_unregister(nv_ctl_handle);
#endif
}


/* this is only called when the vmas are duplicated.
 * this appears to only happen when the process is cloned to create
 * a new process, and not when the process is threaded.
 *
 * increment the usage count for the physical pages, so when this
 * clone unmaps the mappings, the pages are not deallocated under
 * the original process.
 */
void
nv_kern_vma_open(struct vm_area_struct *vma)
{
    MOD_INC_USE_COUNT;
}


void
nv_kern_vma_release(struct vm_area_struct *vma)
{
    MOD_DEC_USE_COUNT;
}

#if defined(IA64) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 9))

/* AGP allocations under the 460GX are not mapped to the aperture
 * addresses by the CPU.  This nopage handler will fault on CPU
 * accesses to AGP memory and map the address to the correct page.
 */
struct page *nv_kern_vma_nopage(struct vm_area_struct *vma, unsigned long address, int write_access)
{
    nv_alloc_t *at;
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_FILEP(vma->vm_file);
    struct page *page_ptr;
    int rm_status;

    nvl = NV_GET_NVL_FROM_FILEP(vma->vm_file);
    if (nvl == NULL)
        return NOPAGE_SIGBUS;
    at = nvl_find_agp_alloc_by_vma(nvl, (unsigned long) vma->vm_start);
    if (at == NULL)
        return NOPAGE_SIGBUS;
    rm_status = KernMapAGPNopage((void *)address, vma, at->priv_data, (void **)&page_ptr);
    if (rm_status)
        return NOPAGE_SIGBUS;
    return page_ptr;
}
#endif

struct vm_operations_struct nv_vm_ops = {
    nv_kern_vma_open,
    nv_kern_vma_release,  /* "close" */
#if defined(IA64) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 9))
    nv_kern_vma_nopage,
#endif
};


/*
** nv_kern_open
**
** nv driver open entry point.  Sessions are created here.
*/
int nv_kern_open(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv = (nv_state_t *) 0;
    nv_linux_state_t *nvl = (nv_linux_state_t *) 0;
    nv_file_private_t *nvfp;
    int devnum;
    int rc, status;

    /* Grab a file private area and save 'nv' in file structure */
    NV_KMALLOC(nvfp, sizeof(nv_file_private_t));
    if ( ! nvfp)
        return -ENOMEM;
    memset(nvfp, 0, sizeof(*nvfp));
    file->private_data = nvfp;

    /* for control device, just jump to its open routine */
    /* after setting up the private data */
    if (NV_DEVICE_IS_CONTROL_DEVICE(inode->i_rdev))
        return nv_kern_ctl_open(inode, file);

    /* what device are we talking about? */
    devnum = NV_DEVICE_NUMBER(inode->i_rdev);
    if (devnum >= NV_MAX_DEVICES)
    {
        rc = -ENODEV;
        goto failed;
    }

    MOD_INC_USE_COUNT;

    nvl = &nv_linux_devices[devnum];
    nv = NV_STATE_PTR(nvl);

    NV_DMSG(nv, "open by process %d (current is 0x%x)", current->pid, current);
    nv_lock_ldata(nv);

    NV_HIDE_IN_FILEP(file, nvl);

    /*
     * map the memory and allocate isr on first open
     */

    if ( ! (nv->flags & NV_FLAG_OPEN))
    {
        if (nv->device_id == 0)
        {
            NV_EMSG((nv_state_t *) 0, "open of nonexistent device %d", devnum);
            rc = -ENXIO;
            goto failed;
        }

        if ((rc = nv_map_device(nv)))
        {
            NV_EMSG(nv, "nv_map_device failed");
            goto failed;
        }

        if ( ! RmInitAdapter(nv))
        {
            NV_EMSG(nv, "RmInitAdapter failed");
            rc = -EIO;
            goto failed;
        }
        nv->flags |= NV_FLAG_POSTED;

        status = request_irq(nv->interrupt_line, nv_kern_isr,
                             SA_INTERRUPT | SA_SHIRQ, "nvidia",
                             (void *) nvl);
        if (status != 0)
        {
            if ( nv->interrupt_line && (status == -EBUSY) )
            {
                printk("NV: Tried to get irq %d, but another driver",
                    (unsigned int) nv->interrupt_line);
                printk(" has it and is not sharing it.\n");
                printk("NV: you may want to verify that an audio driver");
                printk(" isn't using the irq\n");
            }
            NV_EMSG(nv, "isr request failed 0x%x", status);
            rc = -EIO;
            goto failed;
        }

#if !defined (KERNEL_2_2)
        NV_VMALLOC(nvl->event_queue, sizeof(struct __wait_queue_head));
        osMemSet(nvl->event_queue, 0, sizeof(struct __wait_queue_head));
        if (nvl->event_queue == NULL)
            goto failed;

        init_waitqueue_head(GET_EVENT_QUEUE(nvl));
#else
        nvl->event_queue = NULL;
#endif

        nv->flags |= NV_FLAG_OPEN;
    }

    nv->usage_count++;
    nv_unlock_ldata(nv);

    return 0;

 failed:
    MOD_DEC_USE_COUNT;
    if (nv)
        nv_unmap_device(nv);
    nv_unlock_ldata(nv);
    return rc;
}


/*
** nv_kern_close
**
** Master driver close entry point.
*/

int nv_kern_close(
    struct inode *inode,
    struct file *file
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_FILEP(file);
    nv_state_t *nv = NV_STATE_PTR(nvl);

    /* for control device, just jump to its open routine */
    /* after setting up the private data */
    if (NV_DEVICE_IS_CONTROL_DEVICE(inode->i_rdev))
        return nv_kern_ctl_close(inode, file);

    NV_DMSG(nv, "close");

    nv_unix_free_all_unused_clients(nv, current->pid, (void *) file);

    nv_lock_ldata(nv);
    if (--nv->usage_count == 0)
    {
        int counter = 0;

        /* turn off interrupts.
        ** be careful to make sure any pending bottom half gets run
        **  or disabled before calling RmShutdownAdapter() since
        **  it will free up the pdev.  This is hard to see on single
        **  cpu systems, but easy on dual cpu :-)
        */
        // nv_interrupts_disable(nv);
        RmDisableAdapter(nv);

        /* give it a moment to allow any bottom half to run */

#define MAX_BH_TASKS 10
        while ((nv->bh_count) && (counter < MAX_BH_TASKS))
        {
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ/50);
            counter++;
        }

        /* free the irq, which may block until any pending interrupts */
        /* are done being processed. */
        free_irq(nv->interrupt_line, (void *) nv);

        nvl->bh->data = (void *) 0;

        RmShutdownAdapter(nv);

        (void) nv_unmap_device(nv);

        /*
         * Make sure we have freed up all the mappings. The kernel
         * should do this automagically before calling close
         */
        if (nvl->alloc_queue)
        {
            if (nvl->alloc_queue->vma)
            {
                NV_EMSG(nv,
                    "still have vm que at nv_close(): 0x%x to 0x%x",
                    nvl->alloc_queue->vma->vm_start,
                    nvl->alloc_queue->vma->vm_end);
            }
        }

#if !defined (KERNEL_2_2)
        /* this only needs to be freed on 2.4 and later kernels */
        NV_VFREE(nvl->event_queue);
        nvl->event_queue = NULL;
#endif

        /* leave INIT flag alone so we don't reinit every time */
        nv->flags &= ~(NV_FLAG_OPEN | NV_FLAG_WAITING);
    }
    nv_unlock_ldata(nv);

    /* free up our per file private data */
    if (file->private_data)
        NV_KFREE(file->private_data);
    file->private_data = (void *) 0;

    MOD_DEC_USE_COUNT;

    return 0;
}

int nv_kern_mmap(
    struct file  *file,
    struct vm_area_struct *vma
)
{
    int pages;
    nv_alloc_t *at;
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_FILEP(file);
    nv_state_t *nv = NV_STATE_PTR(nvl);

    NV_DMSG(nv, "mmap([0x%lx-0x%lx] off=0x%lx)",
        vma->vm_start,
        vma->vm_end,
        LINUX_VMA_OFFS(vma));

    // be a bit paranoid for now
    if ((NV_MASK_OFFSET(LINUX_VMA_OFFS(vma))) ||
        (NV_MASK_OFFSET(vma->vm_start)) ||
        (NV_MASK_OFFSET(vma->vm_end)))
    {
        return -ENXIO;
    }

    pages = (vma->vm_end - vma->vm_start) >> PAGE_SHIFT;

    /* Max is the size of the framebuffer aperture */
    if (pages > nv->fb.size / PAGE_SIZE)
        return -ENOMEM;

    // we have our own version to keep the module count right
    vma->vm_ops = &nv_vm_ops;

    /*
     * figure out the range and map it in
     */


    /* NV reg space */
    if ((LINUX_VMA_OFFS(vma) >= NV_MMAP_REG_OFFSET) &&
        ((LINUX_VMA_OFFS(vma) + (vma->vm_end - vma->vm_start)) <= NV_MMAP_FB_OFFSET))
    {
#if defined(IA64)
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif
        if (remap_page_range(vma->vm_start,
                             (u32)(nv->regs.address) + LINUX_VMA_OFFS(vma) - NV_MMAP_REG_OFFSET,
                             vma->vm_end - vma->vm_start,
                             vma->vm_page_prot))
            return -EAGAIN;

        /* mark it as IO so that we don't dump it on core dump */
        vma->vm_flags |= VM_IO;
    }

    /* NV fb space */
    else if ((LINUX_VMA_OFFS(vma) >= NV_MMAP_FB_OFFSET) &&
             ((LINUX_VMA_OFFS(vma) + (vma->vm_end - vma->vm_start)) <= NV_MMAP_ALLOCATION_OFFSET))
    {
#if defined(IA64)
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif
        if (remap_page_range(vma->vm_start,
                             (u32)(nv->fb.address) + LINUX_VMA_OFFS(vma) - NV_MMAP_FB_OFFSET,
                             vma->vm_end - vma->vm_start,
                             vma->vm_page_prot))
            return -EAGAIN;

        // mark it as IO so that we don't dump it on core dump
        vma->vm_flags |= VM_IO;
    }

    /* Magic allocator */
    else if (LINUX_VMA_OFFS(vma) == NV_MMAP_ALLOCATION_OFFSET)
    {
        unsigned long page = 0, pos, start;
        int i = 0;

        // protect nvl->alloc_queue with global RM lock

        nv_lock_at(nv);
        at = nvl_find_alloc_by_num_pages(nvl, pages, NULL);

        if (at == NULL)
        {
            nv_unlock_at(nv);
            printk("NVRM: couldn't find pre-allocated memory!\n");
            return -EAGAIN;
        }

        at->vma = vma;

        start = vma->vm_start;
        while (pages--)
        {
            page = (unsigned long) at->page_table[i++];
            if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
              	return -EAGAIN;
            start += PAGE_SIZE;
            pos += PAGE_SIZE;
       	}
        nv_unlock_at(nv);

        /* prevent the swapper from swapping it out */
        vma->vm_flags |= VM_LOCKED;
    }

    /* AGP allocator */
    else if (LINUX_VMA_OFFS(vma) == NV_MMAP_AGP_OFFSET)
    {
        // protect nvl->alloc_queue with global RM lock

        nv_lock_at(nv);
        at = nvl_find_alloc_by_num_pages(nvl, pages, NULL);

        if (at == NULL)
        {
            nv_unlock_at(nv);
            printk("NVRM: couldn't find pre-allocated memory!\n");
            return -EAGAIN;
        }

        at->vma = vma;

        if (NV_OSAGP_ENABLED(nv))
        {
#if !defined(IA64) || (LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 9))
            KernMapAGPPages(vma, at->priv_data);
#else
        vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);
#endif
        }
        else
        {
            // Protect structures modified by NvMapAGPPages
            NvMapAGPPages(nv->pdev, (void **) &vma->vm_start,
                          at->class, at->priv_data);

            /* assumes we're not supporting IA64 AGP ourselves */
            at->agp_mapping = (void *)nv_get_phys_address(vma->vm_start);
        }
        nv_unlock_at(nv);
        /* prevent the swapper from swapping it out */
        vma->vm_flags |= VM_LOCKED;
    } else {
        printk("NVRM: unknown memory type\n");
        return -EINVAL;
    }

    vma->vm_file = file;

    /* just increment usage count, rather than calling vma_open */
    MOD_INC_USE_COUNT;

    return 0;
}


unsigned int nv_kern_poll(
    struct file *file,
    poll_table *wait
)
{
    unsigned int mask = 0;
    nv_linux_state_t *nvl;
    nv_state_t *nv;
    nv_file_private_t *nvfp;

    nvfp = file->private_data;
    nvl = NV_GET_NVL_FROM_FILEP(file);
    nv = NV_STATE_PTR(nvl);

    NV_DMSG(NV_STATE_PTR(nvl), "poll()");

    if (nvl->nv_state.device_number == NV_CONTROL_DEVICE_NUMBER)
        return nv_kern_ctl_poll (file, wait);

    // add us to the list
    poll_wait(file, GET_EVENT_QUEUE(nvl), wait);

    nv_lock_ldata(nv);

    // wake the user on any file-specific event, or a general vblank
    if (nvfp->any_fired_notifiers || nvl->vblank_notifier)
    {
        if (nvl->vblank_notifier)
        {
            // don't clear vblank_notifier until everyone's been notified
            nvl->waiting_for_vblank--;
            if (nvl->waiting_for_vblank == 0)
            {
                nvl->vblank_notifier = 0;
                NV_STATE_PTR(nvl)->flags &= ~NV_FLAG_WAITING;
            }
        }

        // tell user some have been posted
        mask |= POLLPRI;
        NV_DMSG(NV_STATE_PTR(nvl), "poll() live on entry!");
    } else {
        // keep track of how many people are waiting for vblank
        nvl->waiting_for_vblank++;
        NV_STATE_PTR(nvl)->flags |= NV_FLAG_WAITING;
    }

    nv_unlock_ldata(nv);

    return mask;
}

//
// nv_kern_ioctl
//
// nv driver ioctl entry point.
//

#ifdef __KERNEL__
#define COPYIN(dst, src, bytes)     copy_from_user((dst), (src), (bytes))
#define COPYOUT(dst, src, bytes)   copy_to_user((dst), (src), (bytes))
#else
#define COPYIN(dst, src, bytes)     memcpy((dst), (src), (bytes)), 0
#define COPYOUT(dst, src, bytes)   memcpy((dst), (src), (bytes)), 0
#endif

#define DO_OR_DIE(n) { if (n) { status = -EFAULT; goto done; } }

/*
 * some ioctl's can only be done on actual device, others only on the control device
 */
#define CTL_DEVICE_ONLY(nv) { if ( ! ((nv)->flags & NV_FLAG_CONTROL)) { status = -EINVAL; goto done; } }

#define ACTUAL_DEVICE_ONLY(nv) { if ((nv)->flags & NV_FLAG_CONTROL) { status = -EINVAL; goto done; } }


/* todo:
   need ioctl to raise a thread priority that is not superuser
       set its priority to SCHED_FIFO which is simple
       priority scheduling w/ disabled timeslicing
*/

int nv_kern_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int cmd,
    unsigned long i_arg)
{
    int size;
    int status = 0;
    nv_linux_state_t *nvl;
    nv_state_t *nv;
    void *arg = (void *) i_arg;

    nvl = NV_GET_NVL_FROM_FILEP(file);
    nv = NV_STATE_PTR(nvl);

    NV_DMSG(nv, "ioctl(0x%x, 0x%x)", _IOC_NR(cmd), (unsigned int) i_arg);

    size = _IOC_SIZE(cmd);

    switch (cmd)
    {
        /* debug tool; zap the module use count so we can unload driver */
        /*             even if it is confused */
        case NV_IOCTL_MODULE_RESET:
            atomic_set(&__this_module.uc.usecount, 1);
            break;

        /* enable/disable interrupts
           used by X driver to make card quiet while Xserver unnecessarily
           diddles around in its PCI config space
         */
        case NV_IOCTL_INTERRUPTS:
        {
            int state = i_arg;
            ACTUAL_DEVICE_ONLY(nv);
            NV_DMSG(nv, "%sabling interrupts", state ? "En" : "Dis");
            if (state)
                nv_interrupts_enable(nv);
            else
                nv_interrupts_disable(nv);
            break;
        }

        case NV_IOCTL_MEMORY_VTOP:
        {
            nv_ioctl_memory_vtop_t params;
            nv_ioctl_memory_vtop_t *userbuf = arg;

            ACTUAL_DEVICE_ONLY(nv);

            /* copy in the user command buffer */
            DO_OR_DIE(COPYIN(&params, userbuf, sizeof(params)));

            status = nv_vtop(&params, 1);

            break;
        }

        /* pass out info about the card */
        case NV_IOCTL_CARD_INFO:
        {
            nv_ioctl_card_info_t ci;
            nv_ioctl_card_info_t *userbuf = arg;
            nv_linux_state_t *tnvl;
            int i;

            CTL_DEVICE_ONLY(nv);

            for (i = 0, tnvl = nv_linux_devices; tnvl < nv_linux_devices + NV_MAX_DEVICES; tnvl++, i++)
            {
                nv_state_t *tnv = NV_STATE_PTR(tnvl);
                (void) memset(&ci, 0, sizeof(ci));
                if (tnv->device_id)
                {
                    ci.flags = NV_IOCTL_CARD_INFO_FLAG_PRESENT;
                    ci.bus = tnv->bus;
                    ci.slot = tnv->slot;
                    ci.vendor_id = tnv->vendor_id;
                    ci.device_id = tnv->device_id;
                    ci.interrupt_line = tnv->interrupt_line;
                    ci.reg_address = tnv->regs.address;
                    ci.reg_size = tnv->regs.size;
                    ci.fb_address = tnv->fb.address;
                    ci.fb_size = tnv->fb.size;
                }
                /* copy the card info out to user */
                /* just copying 0's for nonexistent cards */
                DO_OR_DIE(COPYOUT(userbuf + i, &ci, sizeof(ci)));
            }
            break;
        }

	/* set a card to be primary (not post it) */
	case NV_IOCTL_PRIMARY_CARD:
	{
	    nv_ioctl_primary_card_t params;
	    nv_ioctl_primary_card_t *userbuf = arg;

	    CTL_DEVICE_ONLY(nv);

	    /* copy in the user command buffer */
	    DO_OR_DIE(COPYIN(&params, userbuf, sizeof(params)));

	    status = nvos_set_primary_card(&params);

	    break;
	}

        default:
            status = nvExtEscape(file, nv, cmd, arg) ? 0 : -EINVAL;
            break;
    }

 done:
    return status;
}

/*
 * driver receives an interrupt
 *    if someone waiting, then hand it off.
 */
void nv_kern_isr(
    int   irq,
    void *arg,
    struct pt_regs *regs
)
{
    nv_linux_state_t *nvl = (void *) arg;
    nv_state_t *nv = NV_STATE_PTR(nvl);
    U032 need_to_run_bottom_half = 0;

    RmIsr(nv->device_number, &need_to_run_bottom_half);
    if (need_to_run_bottom_half)
    {
        nv_lock_bh(nv);
        nv->bh_count++;
        nvl->bh->data = nv->pdev;
        queue_task(nvl->bh, &tq_immediate);
        mark_bh(IMMEDIATE_BH);
        nv_unlock_bh(nv);
    }
}

#ifdef CONFIG_PM
/* kernel calls us with a power management event */
int
nv_kern_pm(
    struct pm_dev *dev,
    pm_request_t rqst,
    void *data
)
{
    /*printk("nv_kern_pm event: rqst 0x%x data 0x%lx\n",
        (unsigned int) rqst,
        (unsigned long) data);*/

    switch (rqst)
    {
#if 0
        /* XXX PM HACK!! for now, let's try this */
        nv_linux_state_t *lnv = &nv_linux_devices[0];

        // our video bios doesn't support APM, only ACPI
        // for now, return an error to try and keep the machine
        // from entering suspend/resume, so as to not lose any
        // user's data.
        case PM_RESUME:
            printk("NVRM: received PM resume event\n");
            RmPowerManagement(NV_STATE_PTR(lnv)->pdev, 0, rqst);
            break;

        case PM_SUSPEND:
            printk("NVRM: received PM suspend event\n");
            RmPowerManagement(NV_STATE_PTR(lnv)->pdev, 0, rqst);
            break;
#else
        case PM_RESUME:
        case PM_SUSPEND:
        //    printk("NVRM: avoiding suspend request, don't want to shutdown!!\n");
            return 1;
#endif

        default:
        //    printk("NVRM: received unknown PM event: 0x%x\n", rqst);
            return 1;
    }
    return 0;
}
#endif

/*
** nv_kern_ctl_open
**
** nv control driver open entry point.  Sessions are created here.
*/
int nv_kern_ctl_open(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv;
    nv_linux_state_t *nvl;

    nvl = &nv_ctl_device;
    nv = (nv_state_t *) nvl;

    nv_lock_ldata(nv);
    nv->device_number = NV_CONTROL_DEVICE_NUMBER;

    /* save the nv away in file->private_data */
    NV_HIDE_IN_FILEP(file, nv);

    MOD_INC_USE_COUNT;

    NV_DMSG(nv, "open");

    /* if this is the first time the control device has been opened,
     * allocate the wait queue
     */

    if (! nvl->event_queue) {

#if !defined (KERNEL_2_2)
        NV_VMALLOC(nvl->event_queue, sizeof(struct __wait_queue_head));
        osMemSet(nvl->event_queue, 0, sizeof(struct __wait_queue_head));
        if (nvl->event_queue == NULL)
            return -ENOMEM;

        init_waitqueue_head(GET_EVENT_QUEUE(nvl));
#else
        nvl->event_queue = NULL;
#endif
    }

    nv->flags |= NV_FLAG_OPEN + NV_FLAG_CONTROL;

    /* turn off the hotkey occurred bit */

    nv->flags &= ~NV_FLAG_HOTKEY_OCCURRED;

    nv->usage_count++;
    nv_unlock_ldata(nv);

    return 0;
}


/*
** nv_kern_ctl_close
*/
int nv_kern_ctl_close(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv = NV_GET_NV_FROM_FILEP(file);

    NV_DMSG(nv, "close");

    nv_lock_ldata(nv);
    if (--nv->usage_count == 0)
    {
#if !defined (KERNEL_2_2)
        nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);
        /* this only needs to be freed on 2.4 and later kernels */
        NV_VFREE(nvl->event_queue);
        nvl->event_queue = 0;
#endif
        nv->flags = 0;
    }
    nv_unlock_ldata(nv);

    nv_unix_free_all_unused_clients(nv, current->pid, (void *) file);

    /* free up our per file private data */
    if (file->private_data)
        NV_KFREE(file->private_data);
    file->private_data = (void *) 0;

    MOD_DEC_USE_COUNT;

    return 0;
}



/*
 * nv_kern_ctl_poll() - add the process to the wait queue
 */

unsigned int nv_kern_ctl_poll(
    struct file *file,
    poll_table *wait
)
{
    nv_linux_state_t *nvl;
    nv_state_t *nv;
    unsigned int ret = 0;

    nvl = NV_GET_NVL_FROM_FILEP(file);
    nv = NV_STATE_PTR(nvl);

    if (file->f_flags & O_NONBLOCK)
        return -EAGAIN;

    poll_wait(file, GET_EVENT_QUEUE(nvl), wait);

    nv_lock_ldata(nv);

    if (nv->flags & NV_FLAG_HOTKEY_OCCURRED) {
        NV_DMSG (nv, "a hotkey event has occurred");
        nv->flags &= ~NV_FLAG_HOTKEY_OCCURRED;
        ret = POLLIN | POLLRDNORM;
    }
    nv_unlock_ldata(nv);

    return ret;
}




/*
 * nv_set_hotkey_occurred_flag() - set the hotkey flag and wake up anybody
 * waiting on the wait queue
 */

void nv_set_hotkey_occurred_flag(void)
{
    NV_DMSG (NV_STATE_PTR(&nv_ctl_device),"setting the hotkey occurred flag!");

    nv_lock_ldata(&(nv_ctl_device.nv_state));
    nv_ctl_device.nv_state.flags |= NV_FLAG_HOTKEY_OCCURRED;
    nv_unlock_ldata(&(nv_ctl_device.nv_state));

    wake_up_interruptible(GET_EVENT_QUEUE(&nv_ctl_device));
}

struct host_bridge_t {
    unsigned int vendor;
    unsigned int device;
    const char *name;
};

static struct host_bridge_t known_hosts[]  = {
    { PCI_VENDOR_ID_AL,     0x1541,  "ALi M1541            " },
    { PCI_VENDOR_ID_AL,     0x1621,  "ALi M1621            " },
    { PCI_VENDOR_ID_AL,     0x1631,  "ALi M1631            " },
    { PCI_VENDOR_ID_AL,     0x1632,  "ALi M1632            " },
    { PCI_VENDOR_ID_AL,     0x1641,  "ALi M1641            " },
    { PCI_VENDOR_ID_AL,     0x1647,  "ALi M1647            " },
    { PCI_VENDOR_ID_AL,     0x1651,  "ALi M1651            " },
    { PCI_VENDOR_ID_AL,     0x0000,  "Ali (unknown)        " },
    { PCI_VENDOR_ID_AMD,    0x7006,  "AMD Irongate         " },
    { PCI_VENDOR_ID_AMD,    0x700e,  "AMD 761              " },
    { PCI_VENDOR_ID_AMD,    0x700c,  "AMD 760MP            " },
    { PCI_VENDOR_ID_AMD,    0x0000,  "AMD (unknown)        " },
    { PCI_VENDOR_ID_INTEL,  0x7180,  "Intel 440LX          " },
    { PCI_VENDOR_ID_INTEL,  0x7190,  "Intel 440BX          " },
    { PCI_VENDOR_ID_INTEL,  0x71a0,  "Intel 440GX          " },
    { PCI_VENDOR_ID_INTEL,  0x84ea,  "Intel 460GX          " },
    { PCI_VENDOR_ID_INTEL,  0x7120,  "Intel i810           " },
    { PCI_VENDOR_ID_INTEL,  0x7124,  "Intel i810E          " },
    { PCI_VENDOR_ID_INTEL,  0x1130,  "Intel i815           " },
    { PCI_VENDOR_ID_INTEL,  0x2500,  "Intel i820           " },
    { PCI_VENDOR_ID_INTEL,  0x2501,  "Intel i820           " },
    { PCI_VENDOR_ID_INTEL,  0x3575,  "Intel i830M          " },
    { PCI_VENDOR_ID_INTEL,  0x1a20,  "Intel i840           " },
    { PCI_VENDOR_ID_INTEL,  0x1a21,  "Intel i840           " },
    { PCI_VENDOR_ID_INTEL,  0x1a22,  "Intel i840           " },
    { PCI_VENDOR_ID_INTEL,  0x1a30,  "Intel i845           " },
    { PCI_VENDOR_ID_INTEL,  0x2530,  "Intel i850           " },
    { PCI_VENDOR_ID_INTEL,  0x2531,  "Intel i850           " },
    { PCI_VENDOR_ID_INTEL,  0x2532,  "Intel i860           " },
    { PCI_VENDOR_ID_INTEL,  0x0000,  "Intel (unknown)      " },
    { PCI_VENDOR_ID_NVIDIA, 0x01a4,  "nVidia nForce        " },
    { PCI_VENDOR_ID_NVIDIA, 0x01a5,  "nVidia nForce        " },
    { PCI_VENDOR_ID_NVIDIA, 0x01a6,  "nVidia nForce        " },
    { PCI_VENDOR_ID_NVIDIA, 0x0000,  "nVidia (unknown)     " },
    { PCI_VENDOR_ID_SI,     0x0530,  "SiS 530              " },
    { PCI_VENDOR_ID_SI,     0x0540,  "SiS 540              " },
    { PCI_VENDOR_ID_SI,     0x0550,  "SiS 550              " },
    { PCI_VENDOR_ID_SI,     0x0620,  "SiS 620              " },
    { PCI_VENDOR_ID_SI,     0x0630,  "SiS 630              " },
    { PCI_VENDOR_ID_SI,     0x0645,  "SiS 645              " },
    { PCI_VENDOR_ID_SI,     0x0650,  "SiS 650              " },
    { PCI_VENDOR_ID_SI,     0x0730,  "SiS 730              " },
    { PCI_VENDOR_ID_SI,     0x0735,  "SiS 735              " },
    { PCI_VENDOR_ID_SI,     0x0740,  "SiS 740              " },
    { PCI_VENDOR_ID_SI,     0x0000,  "SiS (unknown)        " },
    { PCI_VENDOR_ID_VIA,    0x0391,  "Via Apollo Pro KX133 " },
    { PCI_VENDOR_ID_VIA,    0x0305,  "Via Apollo Pro KT133 " },
    { PCI_VENDOR_ID_VIA,    0x3099,  "Via Apollo Pro KT266 " },
    { PCI_VENDOR_ID_VIA,    0x0691,  "Via Apollo Pro       " },
    { PCI_VENDOR_ID_VIA,    0x0501,  "Via MVP4             " },
    { PCI_VENDOR_ID_VIA,    0x0597,  "Via VP3              " },
    { PCI_VENDOR_ID_VIA,    0x0598,  "Via MVP3             " },
    { PCI_VENDOR_ID_VIA,    0x0000,  "Via (unknown)        " },
    { 0x0000,               0x0000,  "Unkown Host Bridge   " },
};

int nv_kern_read_cardinfo(char *page, char **start, off_t off,
        int count, int *eof, void *data)
{
    struct pci_dev *dev;
    char device[25];
    char *type, *fmt, *tmpstr;
    int bios[4], len = 0, i;
    U032 video_bios_version;

    nv_state_t *nv;
    nv = (nv_state_t *) data;

    dev = nv_find_pci_dev(nv);
    if (!dev)
        return 0;
    
    /* sChipInfoChar and CHIP_INFO_CHAR is defined in nv-ids.h */

    tmpstr = "Unknown";
    for (i = 0; i < sizeof(sChipInfoChar) / sizeof(CHIP_INFO_CHAR); i++) {
        if (sChipInfoChar[i].ulDevID == dev->device) {
            tmpstr = sChipInfoChar[i].psChip;
            break;
        }
    }
    sprintf(device, "%s", tmpstr);
    
    len += sprintf(page+len, "Model: \t\t %s\n", device);
    len += sprintf(page+len, "IRQ:   \t\t %d\n", nv->interrupt_line);

    video_bios_version = RmGetVideoBiosVersion(nv->pdev);

    if (!video_bios_version) {
        /* before RmInitAdapter */
        len += sprintf(page+len, "Video BIOS: \t ??.??.??.??\n");
    } else {
        bios[0] = (video_bios_version & 0xff000000) >> 24;
        bios[1] = (video_bios_version & 0x00ff0000) >> 16;
        bios[2] = (video_bios_version & 0x0000ff00) >>  8;
        bios[3] = (video_bios_version & 0x000000ff) >>  0;

        fmt = "Video BIOS: \t %02x.%02x.%02x.%02x\n";
        len += sprintf(page+len, fmt, bios[0], bios[1], bios[2], bios[3]);
    }

    type = nvos_find_agp_capability(dev) ? "AGP" : "PCI";
    len += sprintf(page+len, "Card Type: \t %s\n", type);

    return len;
}

int nv_kern_read_version(char *page, char **start, off_t off,
        int count, int *eof, void *data)
{
    int len = 0;
    
    len += sprintf(page+len, "NVRM version: %s\n", pNVRM_ID);
    len += sprintf(page+len, "GCC version:  %s\n", NV_COMPILER);
    
    return len;
}

int nv_kern_read_agpinfo(char *page, char **start, off_t off,
        int count, int *eof, void *data)
{
    struct pci_dev *dev;
    struct host_bridge_t *host;
    char   *fw, *sba;
    u8     cap_ptr;
    u32    status, command;
    int    len = 0;
    
    nv_state_t *nv;
    nv = (nv_state_t *) data;

    if (nv) {
        dev = nv_find_pci_dev(nv);
        if (!dev)
            return 0;
    } else {
        dev = nvos_find_agp_by_class(PCI_CLASS_BRIDGE_HOST);
        if (!dev)
            return 0;

        host = known_hosts;
        do {
            if (dev->vendor != host->vendor && host->vendor != 0) {
                host++;
                continue;
            }
            if (dev->device == host->device || host->device == 0) {
                len += sprintf(page, "Host Bridge: \t %s\n", host->name);
                break;
            }
            host++;
        } while (1);
    }

    /* what can this AGP device do? */
    cap_ptr = nvos_find_agp_capability(dev);

    pci_read_config_dword(dev, cap_ptr + 4, &status);
    pci_read_config_dword(dev, cap_ptr + 8, &command);

    fw  = (status & 0x00000010) ? "Supported" : "Not Supported";
    sba = (status & 0x00000200) ? "Supported" : "Not Supported";

    len += sprintf(page+len, "Fast Writes: \t %s\n", fw);
    len += sprintf(page+len, "SBA: \t\t %s\n", sba);

    len += sprintf(page+len, "AGP Rates: \t %s%s%s\n",
            (status & 0x00000004) ? "4x " : "",
            (status & 0x00000002) ? "2x " : "",
            (status & 0x00000001) ? "1x " : "");

    len += sprintf(page+len, "Registers: \t 0x%08x:0x%08x\n", status, command);

    return len;
}

int nv_kern_read_status(char *page, char **start, off_t off,
        int count, int *eof, void *data)
{
    struct pci_dev *dev;
    char   *fw, *sba, *drv;
    int    len = 0;
    u8     cap_ptr;
    u32    scratch;
    u32    status, command;

    nv_state_t *nv;
    nv = (nv_state_t *) data;

    dev = nvos_find_agp_by_class(PCI_CLASS_BRIDGE_HOST);
    if (!dev)
        return 0;
    cap_ptr = nvos_find_agp_capability(dev);

    pci_read_config_dword(dev, cap_ptr + 4, &status);
    pci_read_config_dword(dev, cap_ptr + 8, &command);

    dev = nvos_find_agp_by_class(PCI_CLASS_DISPLAY_VGA);
    if (!dev)
        return 0;
    cap_ptr = nvos_find_agp_capability(dev);

    pci_read_config_dword(dev, cap_ptr + 4, &scratch);
    status &= scratch;
    pci_read_config_dword(dev, cap_ptr + 8, &scratch);
    command &= scratch;

    if (command & 0x100) {
        len += sprintf(page+len, "Status: \t Enabled\n");

        drv = NV_OSAGP_ENABLED(nv) ? "AGPGART" : "NVIDIA";
        if (!NV_AGP_ENABLED(nv)) drv = "AGPGART (inactive)";
        len += sprintf(page+len, "Driver: \t %s\n", drv);

        len += sprintf(page+len, "AGP Rate: \t %dx\n", command & 0x7);

        fw = (command & 0x00000010) ? "Enabled" : "Disabled";
        len += sprintf(page+len, "Fast Writes: \t %s\n", fw);

        sba = (command & 0x00000200) ? "Enabled" : "Disabled";
        len += sprintf(page+len, "SBA: \t\t %s\n", sba);
    } else {
        len += sprintf(page+len, "Status: \t Disabled\n");
    }

    return len;
}

int nv_kern_read_legacy(char *page, char **start, off_t off,
        int count, int *eof, void *data)
{
    char *txt = "This file has been deprecated. What you are looking for \n"
                "can now be found in /proc/driver/nvidia. Please see the \n"
                "README for more information on the new /proc interface. \n";

    return sprintf(page, txt);
}


/***
 *** EXPORTS to rest of resman
 ***/

/*
 * Convert virtual (user or kernel) to "physical address"
 *   Here physical means the bus address, as we're just prepping
 *   to give it to dma
 */
int nv_vtop(
    nv_ioctl_memory_vtop_t *params,
    int                     copy_to_user
)
{
    void *v;
    int l;
    unsigned long p;		/* physical addr */
    u32 *physical_addresses;

    // page align everything

    v = NV_ALIGN_PTR_DOWN(params->buffer, NVHW_PAGE_SIZE);
    l = NV_ALIGN_UP((unsigned long)params->byte_length +
                    ((unsigned long)params->buffer - (unsigned long)v), NVHW_PAGE_SIZE);

    for (physical_addresses = params->physical_addresses;
         l > 0;
         physical_addresses++)
    {
        // get the physical address and copy it out to caller's buffer
        p = (unsigned long) nv_get_phys_address((unsigned long) v);

        NV_DMSG((nv_state_t *) 0, "vtop: 0x%p to p 0x%08x", v, (unsigned int) p);
        if (copy_to_user)
            put_user(p, physical_addresses);
        else
            *physical_addresses = p;
        l -= PAGE_SIZE;
        v = (void *)((unsigned long)v + PAGE_SIZE);
    }

    return 0;
}

void *nv_find_kernel_mapping(
    nv_state_t    *nv,
    void          *address
)
{
    nv_alloc_t *at;
    at = nvl_find_alloc(NV_GET_NVL_FROM_NV_STATE(nv), (unsigned long) address,
        (nv_alloc_t **) 0);
    if (at && at->page_table)
    {
        unsigned int page = ((unsigned long) address - at->vma->vm_start) >>
                            PAGE_SHIFT;
        return __va(at->page_table[page]);
    }

    return (void *) 0;
}

#if defined(IA64)
#  define KERN_PAGE_MASK      _PFN_MASK
#else
#  define KERN_PAGE_MASK      PAGE_MASK
#endif

/* virtual address to physical page address */
unsigned long
nv_get_phys_address(unsigned long address)
{
    pgd_t *pg_dir;
    pmd_t *pg_mid_dir;
    pte_t *pg_table;

    /* XXX do we really need this? */
    if (address > VMALLOC_START)
        address = VMALLOC_VMADDR(address);

#if defined(IA64)
    if (address > __IA64_UNCACHED_OFFSET)
        return address;
#endif

    /* direct-mapped kernel address */
    /* is this ok for IA64? */
    if ((address > PAGE_OFFSET) && (address < VMALLOC_START))
        return __pa(address);

    if (address > VMALLOC_START)
       pg_dir = pgd_offset_k(address);
    else
       pg_dir = pgd_offset(current->mm, address);

    if (pgd_none(*pg_dir))
        goto failed;

    pg_mid_dir = pmd_offset(pg_dir, address);
    if (pmd_none(*pg_mid_dir))
        goto failed;

    pg_table = pte_offset(pg_mid_dir, address);
    if (!pte_present(*pg_table))
        goto failed;

    return ((pte_val(*pg_table) & KERN_PAGE_MASK) | NV_MASK_OFFSET(address));

  failed:
    return (unsigned long) NULL;
}


/* at the 'at' to our list */
#define NV_ADD_AT(nv, at) \
    { \
        nv_lock_at(NV_STATE_PTR(nv)); \
        at->next = nvl->alloc_queue; \
        nvl->alloc_queue = at; \
        nv_unlock_at(NV_STATE_PTR(nv)); \
    }

#define NV_REMOVE_AT(nv, at) \
    { \
        nv_lock_at(NV_STATE_PTR(nv)); \
        nvl->alloc_queue = at->next; \
        at->next = NULL; \
        nv_unlock_at(NV_STATE_PTR(nv)); \
    }

/* allocate memory for DMA push buffers */
int
nv_alloc_pages(
    nv_state_t *nv,
    void **pAddress,
    unsigned int page_count,
    unsigned int agp_memory,
    unsigned int class,
    void **priv_data
)
{
    nv_alloc_t *at;
    int rm_status = 0;
    nv_linux_state_t *nvl = (nv_linux_state_t *) nv;

    NV_KMALLOC(at, sizeof(nv_alloc_t));
    if (at == NULL)
        return RM_ERROR;

    osMemSet(at, 0, sizeof(nv_alloc_t));

    page_count = NVHW_PAGES_TO_OS_PAGES(page_count);
    at->num_pages = page_count;

    at->process_id = current->pid;
#if !defined (KERNEL_2_2)
    at->thread_gid = current->tgid;
#else
    at->thread_gid = -1;
#endif

    at->class = class;
    at->vma = NULL;

    if (at->class == NV01_ROOT)
    {
        int order, i;
        unsigned long addr;

        calc_order(at->num_pages * PAGE_SIZE, order);
        addr = __get_free_pages(GFP_KERNEL, order);

        if (addr == 0x0)
            goto failed;

        *pAddress = (void*) addr;
        at->direct_kmapping = *pAddress;

        for (i = 0; i < at->num_pages; i++)
        {
            mem_map_reserve(GET_MAP_NR(__pa(addr)));
            addr += PAGE_SIZE;
        }

        NV_ADD_AT(nvl, at);
        return RM_OK;
    }

    if (agp_memory)
    {
        if (NV_AGP_DISABLED(nv))
        {
            goto failed;
        }

        /* allocate agp-able memory */
        if (NV_OSAGP_ENABLED(nv))
        {
            /* agpgart will allocate all of the underlying memory */
            rm_status = KernAllocAGPPages(pAddress, page_count, priv_data);
            if (rm_status)
                goto failed;

            at->agp_mapping = *pAddress;
            at->priv_data = *priv_data;
            NV_ADD_AT(nvl, at);
        } else {
            /* use nvidia's nvagp support */
            at->page_table = nvos_malloc(page_count << PAGE_SHIFT);
            if (at->page_table == NULL)
                goto failed;

            *pAddress = (void *) at->page_table;
            at->class = class;

            /* the 'at' needs to be added before the alloc agp pages call */
            NV_ADD_AT(nvl, at);
            rm_status = NvAllocAGPPages(nv->pdev,
                                        pAddress,
                                        page_count,
                                        class,
                                        priv_data);
            if (rm_status)
            {
                NV_REMOVE_AT(nvl, at);
                goto failed;
            }
            at->priv_data = *priv_data;
        }
        nv->agp_buffers++;
    } else {
        /* allocate general system memory */
        at->page_table = nvos_malloc(page_count << PAGE_SHIFT);
        if (at->page_table == NULL)
            goto failed;

        *pAddress = (void *) at->page_table;
        NV_ADD_AT(nvl, at);
    }

    return rm_status;

failed:
    /* free any pages we may have allocated */
    if (at->page_table)
        nvos_free(at->page_table);

    /* free it */
    NV_KFREE(at);
    return -1;
}

#define NV_REMOVE_AT_FROM_LIST(nvl, at, prev) \
    if (prev) prev->next = at->next; \
    else nvl->alloc_queue = at->next;

#define NV_FAILED_TO_FIND_AT(nv, paddr) \
    { \
        nv_unlock_at(nv); \
        printk("NVRM: couldn't find alloc for 0x%p\n", *paddr); \
        return -1; \
    }

int
nv_free_pages(
    nv_state_t *nv,
    void **pAddress,
    unsigned int page_count,
    unsigned int agp_memory,
    void **priv_data
)
{
    int rmStatus = 0;
    nv_alloc_t *at, *prev;
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

    if (agp_memory)
    {
        /* only lock ldata while removing 'at' from the list */
        nv_lock_at(nv);
        at = nvl_find_agp_alloc(nvl, (unsigned long) *pAddress, &prev);
        if (at == NULL)
            NV_FAILED_TO_FIND_AT(nv, pAddress);
        NV_REMOVE_AT_FROM_LIST(nvl, at, prev);
        nv_unlock_at(nv);

        if (NV_OSAGP_ENABLED(nv))
        {
            rmStatus = KernFreeAGPPages(pAddress, priv_data);
        } else {
            rmStatus = NvFreeAGPPages(nv->pdev, pAddress, priv_data);
            if (rmStatus == 0x0)
                nvos_free(at->page_table);
        }

        /* we may hold off on disabling agp until all buffers are freed */
        if (rmStatus == 0x0)
        {
            nv->agp_buffers--;
            if (!nv->agp_buffers && nv->agp_teardown)
                nv_agp_teardown(nv);
        }
    } else {
        /* only lock ldata while removing 'at' from the list */
        nv_lock_at(nv);
        at = nvl_find_alloc(nvl, (unsigned long) *pAddress, &prev);
        if (at == NULL)
            NV_FAILED_TO_FIND_AT(nv, pAddress);
        NV_REMOVE_AT_FROM_LIST(nvl, at, prev);
        nv_unlock_at(nv);

        if (at->class == NV01_ROOT)
        {
            int order, i;
            unsigned long addr;

            addr = (unsigned long) *pAddress;
            for (i = 0; i < at->num_pages; i++)
            {
                mem_map_unreserve(GET_MAP_NR(__pa(addr)));
                addr += PAGE_SIZE;
            }

            calc_order(at->num_pages * PAGE_SIZE, order);
            free_pages((unsigned long) *pAddress, order);
        }
        else
        {
            nvos_free(at->page_table);
        }
    }

    NV_KFREE(at);
    return rmStatus;
}


#if defined(NV_DEBUG_LOCKS)
#define DEBUG_NONE        0
#define DEBUG_DEADLOCKS   1
#define DEBUG_LOCKING     2
#define DEBUG_ISR_LOCKING 3
static int locking_debuglevel = DEBUG_NONE;
#endif
static void nv_lock_init_locks
( 
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

    /* fix a compiler warning when spin_lock_init() is macroed away
     * and nvl isn't used */
    if (nvl == NULL)
        return;

    spin_lock_init(&nvl->rm_lock);
    spin_lock_init(&nvl->ldata_lock);
    spin_lock_init(&nvl->at_lock);
    spin_lock_init(&nvl->bh_lock);

#if defined(NV_DEBUG_LOCKS)
    nvl->rm_locked     = 0;
    nvl->ldata_locked  = 0;
    nvl->at_locked     = 0;
    nvl->bh_locked     = 0;
    nvl->rm_waiting    = 0;
    nvl->ldata_waiting = 0;
    nvl->at_waiting    = 0;
#endif
}

void nv_lock_rm(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

#if defined(NV_DEBUG_LOCKS)
    int me = (in_interrupt() ? (in_irq() ? -1 : -2) : current->pid);
    int me_waiting = 0;

    // make sure we want to debug locks..
    if (locking_debuglevel && nvl->rm_locked) // &&
        // if we're not debugging ISRs, make sure we're not in an interrupt
        // ((locking_debuglevel < DEBUG_ISR_LOCKING) ? !in_interrupt() : TRUE))
    {
        nv_printf("(%d): %d waiting on rm (%d)\n", current->processor, me,
            nvl->rm_locked);
        if ((nvl->rm_locked == current->pid) && (!in_interrupt()))
        {
            nv_printf("  I already own the lock!\n");
        }
        if (nvl->rm_locked == -1)
            nv_printf("  ISR owns the lock!!\n");
        if (nvl->rm_locked == -2)
            nv_printf("  BH owns the lock!!\n");
        if (in_interrupt())
            nv_printf("  I'm an ISR!!!\n");
        me_waiting = 1;
        nvl->rm_waiting++;
    }
#endif

    spin_lock_irqsave(&nvl->rm_lock, nvl->rm_pflags);

#if defined(NV_DEBUG_LOCKS)
    nvl->rm_locked = me;
    if (locking_debuglevel)
    {
        if (locking_debuglevel > DEBUG_DEADLOCKS)
        {
            if ((locking_debuglevel < DEBUG_ISR_LOCKING) ? !in_interrupt() : TRUE)
                nv_printf("(%d): %d rm\n", current->processor, nvl->rm_locked);
        }
        else
        {
            if (me_waiting)
            {
                nv_printf("   Got it!!\n");
                nvl->rm_waiting--;
            }
        }
    }
#endif
}

void nv_unlock_rm(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);
#if defined(NV_DEBUG_LOCKS)
    if (locking_debuglevel)
    {
        if (nvl->rm_waiting ||
            ((locking_debuglevel > DEBUG_DEADLOCKS) &&
             ((locking_debuglevel < DEBUG_ISR_LOCKING) ? !in_interrupt() : TRUE)))
                nv_printf("(%d): %d drop rm\n", current->processor, nvl->rm_locked);
    }
    nvl->rm_locked = 0;
#endif
    spin_unlock_irqrestore(&nvl->rm_lock, nvl->rm_pflags);
}

static void nv_lock_ldata(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

#if defined(NV_DEBUG_LOCKS)
    int me = (in_interrupt() ? (in_irq() ? -1 : -2) : current->pid);
    int me_waiting = 0;

    // make sure we want to debug locks..
    if (locking_debuglevel && nvl->ldata_locked) // &&
        // if we're not debugging ISRs, make sure we're not in an interrupt
        // ((locking_debuglevel < DEBUG_ISR_LOCKING) ? !in_interrupt() : TRUE))
    {
        nv_printf("(%d): %d wait on ldata (%d)\n", current->processor, me,
            nvl->ldata_locked);
        if (nvl->ldata_locked == current->pid)
            nv_printf("  I already own the lock!\n");
        if (nvl->ldata_locked == -1)
            nv_printf("  ISR owns the lock!!\n");
        if (nvl->ldata_locked == -2)
            nv_printf("  BH owns the lock!!\n");
        me_waiting = 1;
        nvl->ldata_waiting++;
    }
#endif

    spin_lock_irqsave(&nvl->ldata_lock, nvl->ldata_pflags);

#if defined(NV_DEBUG_LOCKS)
    nvl->ldata_locked = me;

    if (locking_debuglevel)
    {
        if (locking_debuglevel > DEBUG_DEADLOCKS)
        {
            if ((locking_debuglevel < DEBUG_ISR_LOCKING) ? !in_interrupt() : TRUE)
                nv_printf("(%d): %d ldata\n", current->processor, nvl->ldata_locked);
        }
        else
            if (me_waiting)
            {
                nv_printf("   Got it!!\n");
                nvl->ldata_waiting--;
            }
    }
#endif
}

static void nv_unlock_ldata(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

#if defined(NV_DEBUG_LOCKS)
    if (locking_debuglevel)
    {
        if (nvl->ldata_waiting ||
            ((locking_debuglevel > DEBUG_DEADLOCKS) &&
             ((locking_debuglevel < DEBUG_ISR_LOCKING) ? !in_interrupt() : TRUE)))
                nv_printf("(%d): %d drop ldata\n", current->processor, nvl->ldata_locked);
    }
    nvl->ldata_locked = 0;
#endif

    spin_unlock_irqrestore(&nvl->ldata_lock, nvl->ldata_pflags);
}


static void nv_lock_at(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

#if defined(NV_DEBUG_LOCKS)
    int me = (in_interrupt() ? (in_irq() ? -1 : -2) : current->pid);
    int me_waiting = 0;

    // make sure we want to debug locks..
    if (locking_debuglevel && nvl->at_locked) // &&
        // if we're not debugging ISRs, make sure we're not in an interrupt
        // ((locking_debuglevel < DEBUG_ISR_LOCKING) ? !in_interrupt() : TRUE))
    {
        nv_printf("(%d): %d wait on at (%d)\n", current->processor, me,
            nvl->at_locked);
        if (nvl->at_locked == current->pid)
            nv_printf("  I already own the lock!\n");
        if (nvl->at_locked == -1)
            nv_printf("  ISR owns the lock!!\n");
        if (nvl->at_locked == -2)
            nv_printf("  BH owns the lock!!\n");
        me_waiting = 1;
        nvl->at_waiting++;
    }
#endif

    spin_lock_irqsave(&nvl->at_lock, nvl->at_pflags);

#if defined(NV_DEBUG_LOCKS)
    nvl->at_locked = me;

    if (locking_debuglevel)
    {
        if (locking_debuglevel > DEBUG_DEADLOCKS)
        {
            if ((locking_debuglevel < DEBUG_ISR_LOCKING) ? !in_interrupt() : TRUE)
                nv_printf("(%d): %d at\n", current->processor, nvl->at_locked);
        }
        else
            if (me_waiting)
            {
                nv_printf("   Got it!!\n");
                nvl->at_waiting--;
            }
    }
#endif
}

static void nv_unlock_at(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

#if defined(NV_DEBUG_LOCKS)
    if (locking_debuglevel)
    {
        if (nvl->at_waiting ||
            ((locking_debuglevel > DEBUG_DEADLOCKS) &&
             ((locking_debuglevel < DEBUG_ISR_LOCKING) ? !in_interrupt() : TRUE)))
                nv_printf("(%d): %d drop at\n", current->processor, nvl->at_locked);
    }
    nvl->at_locked = 0;
#endif

    spin_unlock_irqrestore(&nvl->at_lock, nvl->at_pflags);
}


void nv_lock_bh(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

#if defined(NV_DEBUG_LOCKS)
    int me = (in_irq() ? -1 : -2 );
    int bh_waiting = 0;

    if (locking_debuglevel && nvl->bh_locked)
    {
        nv_printf("(%d) %d wait on bh\n", current->processor, me);
        if (nvl->bh_locked == -1)
            nv_printf("  ISR owns the lock!!\n");
        if (nvl->bh_locked == -2)
            nv_printf("  BH owns the lock!!\n");
        bh_waiting = 1;
    }
#endif

    spin_lock_irqsave(&nvl->bh_lock, nvl->bh_pflags);

#if defined(NV_DEBUG_LOCKS)
    nvl->bh_locked = me;

    if (locking_debuglevel)
    {
        if (locking_debuglevel > DEBUG_DEADLOCKS)
            nv_printf("(%d): %d bh\n", current->processor, nvl->bh_locked);
        else
            if (bh_waiting)
                nv_printf("   Got it!!\n");
    }
#endif
}

void nv_unlock_bh(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

#if defined(NV_DEBUG_LOCKS)
    if (locking_debuglevel >= DEBUG_ISR_LOCKING)
        nv_printf("(%d): %d drop bh\n", current->processor, nvl->bh_locked);
    nvl->bh_locked = 0;
#endif

    spin_unlock_irqrestore(&nvl->bh_lock, nvl->bh_pflags);
}



/*
** post the event
** Called from osNotifyEvent()
**
** We already hold the spinlock when we get in here...
*/
void nv_post_event(
    void     *file_void
)
{
    nv_linux_state_t *nvl;
    nv_file_private_t *nvfp;
    struct file *file = file_void;

    nvfp = file->private_data;
    nvl = NV_GET_NVL_FROM_FILEP(file);

    NV_DMSG(NV_STATE_PTR(nvl), "post");

    nvfp->any_fired_notifiers++;

    wake_up_interruptible(GET_EVENT_QUEUE(nvl));
}

/*
** post vblank specifically
** Called from osVblankNotify(), from VBlank()
*/
void nv_post_vblank(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = (nv_linux_state_t *) nv->os_state;
    NV_DMSG(nv, "post");

    if (nvl->waiting_for_vblank)
        nvl->vblank_notifier++;

    wake_up_interruptible(GET_EVENT_QUEUE(nvl));
}


int
nv_agp_init(
    nv_state_t *nv,
    VOID **phys_start,
    VOID **linear_start,
    VOID *agp_limit,
    U032 config         /* passed in from XF86Config file */
)
{
    U032 status = 1;

    nv_lock_ldata(nv);
    if (config & NVOS_AGP_CONFIG_OSAGP)
    {
        status = KernInitAGP(phys_start, linear_start, agp_limit);

        /* if enabling agpgart was successfull, register it,
         * and check about overrides
         */
        if (status == 0)
        {
            nv->agp_config = NVOS_AGP_CONFIG_OSAGP;
            /* make sure we apply our overrides in this case */
            NvUpdateAGPConfig(nv->pdev);
        }

        /* if agpgart is loaded, but we failed to initialize it,
         * we'd better not attempt nvagp, or we're likely to lock
         * the machine.
         */
        if (status < 0)
        {
            nv_unlock_ldata(nv);
            return 1;
        }
    }

    /* we're either explicitly not using agpgart,
     * or trying to use agpgart failed
     * make sure the user did not specify "use agpgart only"
     */
    if ( (NV_AGP_DISABLED(nv)) && (config & NVOS_AGP_CONFIG_NVAGP) )
    {
        /* make sure the user does not have agpgart loaded */
#if !defined (KERNEL_2_2)
        if (inter_module_get("drm_agp")) {
            inter_module_put("drm_agp");
#else
        if (GET_MODULE_SYMBOL(0, __MODULE_STRING(agp_enable))) {
#endif
            printk("NVRM: not using NVAGP, AGPGART is loaded!!\n");
        } else
            status = RmInitAGP(nv->pdev);
            if (status == RM_OK)
                nv->agp_config = NVOS_AGP_CONFIG_NVAGP;
    }

    nv_unlock_ldata(nv);
    return status;
}

int
nv_agp_teardown(
    nv_state_t *nv
)
{
    U032 status = 1;

    /* little sanity check won't hurt */
    if (NV_AGP_DISABLED(nv))
        return -1;

    nv_lock_ldata(nv);

    /* if agp buffers are still in use, don't teardown just yet */
    if (nv->agp_buffers)
    {
        nv->agp_teardown = 1;
        nv_unlock_ldata(nv);
        return 0;
    }

    if (NV_OSAGP_ENABLED(nv))
    {
        status = KernTeardownAGP();
    } else
        status = RmTeardownAGP(nv->pdev);

    nv->agp_teardown = 0;
    nv->agp_config = NVOS_AGP_CONFIG_DISABLE_AGP;
    nv_unlock_ldata(nv);

    return status;
}

/*
 * this takes a virtual address related to agp and returns the physical address
 * this part is a bit hacky.. nvagp expects a kernel-linear address mapping,
 * but we don't have that. So, what's really passed to us is a "fake linear"
 * address, based at our page_table. nvl_find_agp_alloc() knows this, and will
 * return the 'at' that has the page_table. From there, we deconstruct this
 * 'fake address' to get a "page index" into the virtual address. We use this
 * index to grab a page from our page_table and return that physical address.
 * (if the linear mapping was real, this is the page that would have been
 * addressed by the virtual address. I know, confusing).
 */
int
nv_agp_translate_address(
    nv_state_t *nv,
    void       *base,
    U032        index,
    U032       *paddr
)
{
    nv_alloc_t *at;

    /* find the 'at' */
    at = nvl_find_agp_alloc(NV_GET_NVL_FROM_NV_STATE(nv), (unsigned long) base, 0);
    if (at == NULL)
        return -1;

    /* get the physical address of this page */
    *paddr = (U032) ((NV_UINTPTR_T)at->page_table[index]);

    return 0x0;
}

/* For AGP chipsets such as the 460GX, the virtual address is not mapped
 * to the aperture address on the CPU.  This return returns the mapping so
 * that osGetAddressInfo will work.
 */
int
nv_agp_retrieve_mapping(
    nv_state_t    *nv,
    void          *address,
    unsigned long *agp_address
)
{
    nv_alloc_t *at;

#if defined(IA64) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 9))
    at = nvl_find_agp_alloc_by_vma(NV_GET_NVL_FROM_NV_STATE(nv), (unsigned long) address);
#else
    at = nvl_find_agp_alloc(NV_GET_NVL_FROM_NV_STATE(nv), (unsigned long) address, 0);
#endif
    if ((at == NULL) || (at->agp_mapping == NULL))
        return 1;

    *agp_address = (unsigned long)((NV_UINTPTR_T)at->agp_mapping);

    return 0x0;
}

int
nv_int10h_call(
    nv_state_t *nv,
    U032 *eax,
    U032 *ebx,
    U032 *ecx,
    U032 *edx,
    VOID *buffer
)
{

    // hacked override for the time being..
    if ( (*eax == 0x4f14) && (*ebx == 0x0186) )
    {
        *eax = 0x004f;
        *ecx = ((REG_RD32(0x101000)>>16)&0xf);
    }

    return 0x0;
}
